"""The pyvelop module."""

# region #-- imports --#
from typing import Optional

# endregion


def signal_strength_to_text(rssi: Optional[int]) -> Optional[str]:
    """Convert the given RSSI value to a textual representation."""
    ret: Optional[str] = None
    if rssi is not None:
        if rssi <= 0:
            ret = "Excellent"
        if rssi <= -50:
            ret = "Good"
        if rssi <= -60:
            ret = "Fair"
        if rssi <= -70:
            ret = "Weak"

    return ret



"""Represents a node in the mesh."""

# region #-- imports --#
from __future__ import annotations

from typing import List, Optional

from . import signal_strength_to_text
from .base import MeshDevice

# endregion

NODE_TYPE_PRIMARY = "primary"
NODE_TYPE_SECONDARY = "secondary"


class Node(MeshDevice):
    """Representation of a node in the mesh.  A node provides the connectivity for a device."""

    def __init__(self, **kwargs):
        """Initialise the Node.

        All supplied are arguments are deemed as attributes and are stored in a private variable.

        :param kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.__connected_devices: Optional[List] = None
        self.__parent_name: Optional[str] = None

    @property
    def backhaul(self) -> dict:
        """Get details about the backhaul."""
        ret = {}
        backhaul = self._attribs.get("backhaul", {})
        speed_mbps: Optional[float] = None
        try:
            speed_mbps = float(backhaul.get("speedMbps"))
        except (TypeError, ValueError):
            pass
        if backhaul:
            signal_strength_raw: int = backhaul.get("wirelessConnectionInfo", {}).get(
                "stationRSSI"
            )
            ret = {
                "connection": backhaul.get("connectionType"),
                "last_checked": backhaul.get("timestamp"),
                "speed_mbps": speed_mbps,
                "rssi_dbm": signal_strength_raw,
                "signal_strength": signal_strength_to_text(rssi=signal_strength_raw),
            }

        return ret

    @property
    def connected_devices(self) -> List:
        """List of the devices that are connected to the node.

        :return: List of connected devices in alphabetical order sorted by device name
        """
        connected_devices = self.__connected_devices
        return sorted(connected_devices, key=lambda device: device.get("name"))

    @property
    def firmware(self) -> dict:
        """Get the firmware details for the node.

        N.B. The date doesn't seem to correlate to anything that I can see (I would have thought it was a build
        or install time but that doesn't seem to be the case)

        :return: A dictionary containing the firmware version and date
        """
        ret = {}
        if self._attribs.get("unit", {}):
            ret["version"] = self._attribs.get("unit", {}).get("firmwareVersion")
            ret["date"] = self._attribs.get("unit", {}).get("firmwareDate")
        available_updates = self._attribs.get("updates", {})
        if isinstance(available_updates, dict):
            available_updates = available_updates.get("availableUpdate", {})
            if available_updates:
                ret["latest_version"] = available_updates["firmwareVersion"]
                ret["latest_date"] = available_updates["firmwareDate"]
            else:
                ret["latest_version"] = ret["version"] if "version" in ret else None
                ret["latest_date"] = ret["date"] if "date" in ret else None
        return ret

    @property
    def hardware_version(self) -> str:
        """Get the hardware version of the node.

        :return: A string containing the hardware version
        """
        return self._attribs.get("model", {}).get("hardwareVersion")

    @property
    def last_update_check(self) -> Optional[str]:
        """Get the last time an update was checked for.

        :return: String containing the last update time as per the API
        """
        ret = self._attribs.get("updates", {}).get("lastSuccessfulCheckTime", None)
        return ret

    @property
    def manufacturer(self) -> str:
        """Get the node manufacturer.

        :return: String containing the name of the manufacturer
        """
        return self._attribs.get("model", {}).get("manufacturer")

    @property
    def model(self) -> str:
        """Get the model of the node.

        :return: A string containing the model
        """
        return self._attribs.get("model", {}).get("modelNumber")

    @property
    def parent_ip(self) -> Optional[str]:
        """Get the IP address of the node that this node is attached to.

        :return: A string containing the IP address of the parent or None if no parent
        """
        return self._attribs.get("backhaul", {}).get("parentIPAddress")

    @property
    def parent_name(self) -> str:
        """Return the parent name."""
        return self.__parent_name

    @property
    def serial(self) -> str:
        """Get the serial number of the node.

        :return: A string containing the serial number
        """
        return self._attribs.get("unit", {}).get("serialNumber")

    @property
    def type(self) -> str:
        """Get the node type.

        The node types are represented as primary or secondary.

        :return: A string containing the node type.
        """
        ret = ""
        native_type = self._attribs.get("nodeType", "").lower()
        if native_type == "master":
            ret = NODE_TYPE_PRIMARY
        elif native_type == "slave":
            ret = NODE_TYPE_SECONDARY
        return ret



"""Decorators."""

# region #-- imports --#
import functools

from pyvelop.exceptions import MeshNeedsGatherDetails

# endregion


def needs_gather_details(func):
    """Ensure that async_gather_details has been executed."""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        """Wrap the required function."""
        if not getattr(self, "_Mesh__gather_details_executed", False):
            raise MeshNeedsGatherDetails from None
        ret = func(self, *args, **kwargs)
        return ret

    return wrapper



"""Interact with the JNAP API."""

# region #-- imports --#
from __future__ import annotations

import asyncio
import base64
import json
import logging
from typing import Any, Dict, List, Optional

import aiohttp

from .const import DEF_REDACT
from .exceptions import (
    MeshBadResponse,
    MeshConnectionError,
    MeshInvalidCredentials,
    MeshInvalidInput,
    MeshInvalidOutput,
    MeshNodeNotPrimary,
    MeshTimeoutError,
)
from .logger import Logger

# endregion

_LOGGER = logging.getLogger(__name__)
_LOGGER_VERBOSE = logging.getLogger(f"{__name__}.verbose")


def jnap_url(target) -> str:
    """Return the URL that should be used for the request.

    :param target: the API host
    :return: string containing the base URL for all JNAP requests
    """
    return f"http://{target}/JNAP/"


class Actions:
    """Represents the available actions."""

    ROOT: str = "http://linksys.com/jnap"

    CHECK_PASSWORD: str = f"{ROOT}/core/CheckAdminPassword"
    DELETE_DEVICE: str = f"{ROOT}/devicelist/DeleteDevice"
    GET_BACKHAUL: str = f"{ROOT}/nodes/diagnostics/GetBackhaulInfo"
    GET_DEVICES: str = f"{ROOT}/devicelist/GetDevices3"
    GET_FIRMWARE_UPDATE_SETTINGS: str = (
        f"{ROOT}/firmwareupdate/GetFirmwareUpdateSettings"
    )
    GET_GUEST_NETWORK_INFO: str = f"{ROOT}/guestnetwork/GetGuestRadioSettings2"
    GET_NETWORK_CONNECTIONS: str = (
        f"{ROOT}/nodes/networkconnections/GetNodesWirelessNetworkConnections"
    )
    GET_PARENTAL_CONTROL_INFO: str = (
        f"{ROOT}/parentalcontrol/GetParentalControlSettings"
    )
    GET_SPEEDTEST_RESULTS: str = f"{ROOT}/healthcheck/GetHealthCheckResults"
    GET_SPEEDTEST_STATUS: str = f"{ROOT}/healthcheck/GetHealthCheckStatus"
    GET_STORAGE_PARTITIONS: str = f"{ROOT}/nodes/storage/GetNodesPartitions"
    GET_STORAGE_SMB_SERVER: str = f"{ROOT}/nodes/storage/GetSMBServerSettings"
    GET_UPDATE_FIRMWARE_STATE: str = (
        f"{ROOT}/nodes/firmwareupdate/GetFirmwareUpdateStatus"
    )
    GET_UPDATE_SETTINGS: str = f"{ROOT}/firmwareupdate/GetFirmwareUpdateSettings"
    GET_WAN_INFO: str = f"{ROOT}/router/GetWANStatus3"
    REBOOT: str = f"{ROOT}/core/Reboot"
    SET_GUEST_NETWORK: str = f"{ROOT}/guestnetwork/SetGuestRadioSettings2"
    SET_PARENTAL_CONTROL_INFO: str = (
        f"{ROOT}/parentalcontrol/SetParentalControlSettings"
    )
    START_SPEEDTEST: str = f"{ROOT}/healthcheck/RunHealthCheck"
    TRANSACTION: str = f"{ROOT}/core/Transaction"
    UPDATE_FIRMWARE: str = f"{ROOT}/nodes/firmwareupdate/UpdateFirmwareNow"


class Defaults:
    """Represents the default payloads required for requests."""

    PAYLOADS: Dict[str, Dict] = {
        Actions.GET_SPEEDTEST_RESULTS: {
            "healthCheckModule": "SpeedTest",
            "includeModuleResults": True,
            "lastNumberOfResults": 1,
        },
    }


class Request:
    """Represents a request for the API."""

    def __init__(
        self,
        action: str,
        password: str,
        target: str,
        payload: Optional[List[Dict] | Dict] = None,
        raise_on_error: bool = True,
        session: Optional[aiohttp.ClientSession] = None,
        username: str = "admin",
    ) -> None:
        """Initialise a request.

        :param action: the JNAP action to carry out
        :param password: the password required to communicate with the target
        :param target: the node to send the request to
        :param payload: the additional configuration to pass along with the action
        :param raise_on_error: raise an error if one is found
        :param session: an existing session to use
        :param username: the username required to communicate with the target
        """
        self._action: str = action
        self._creds: str = base64.b64encode(
            bytes(f"{username}:{password}", "utf-8")
        ).decode("ascii")
        self._log_formatter = Logger(prefix=f"{self.__class__.__name__}.")
        self._payload: Optional[List[Dict] | Dict] = payload
        self._raise_on_error: bool = raise_on_error
        self._session: Optional[
            aiohttp.ClientSession
        ] = session or aiohttp.ClientSession(raise_for_status=True)
        self._target: str = target

        self._jnap_url: str = jnap_url(target=self._target)

    async def execute(self, timeout: int = 10) -> Response:
        """Send the request.

        :param timeout: the timeout in seconds for the request, defaults to 10s
        :return: a Response object representing the returned results
        """
        _LOGGER.debug(self._log_formatter.format("entered"))

        headers: Dict[str, str] = {
            "X-JNAP-Authorization": f"Basic {self._creds}",
            "Content-Type": "application/json; charset=UTF-8",
            "X-JNAP-Action": self._action,
        }

        _LOGGER.debug(
            self._log_formatter.format(
                "URL: %s, Headers: %s, Payload: %s, Timeout: %i"
            ),
            self._jnap_url,
            {
                key: value if key not in ("X-JNAP-Authorization") else DEF_REDACT
                for key, value in headers.items()
            },
            json.dumps(self._payload),
            timeout,
        )

        resp: Optional[aiohttp.ClientResponse] = None
        try:
            resp = await self._session.post(
                url=self._jnap_url,
                headers=headers,
                json=self._payload or {},
                timeout=timeout,
            )
            resp_json = await resp.json()
        except asyncio.TimeoutError as err:
            raise MeshTimeoutError from err
        except (
            aiohttp.ClientConnectionError,
            aiohttp.ClientConnectorError,
            aiohttp.ContentTypeError,
        ):
            raise MeshConnectionError from None
        except json.JSONDecodeError as err:
            _LOGGER.debug(self._log_formatter.format("resp: %s"), resp)
            _LOGGER.error(self._log_formatter.format("%s"), err)
            raise err from None

        _LOGGER_VERBOSE.debug(
            self._log_formatter.format("action: %s --> payload: %s --> response: %s"),
            self.action,
            self.payload,
            resp_json,
        )
        _LOGGER.debug(self._log_formatter.format("exited"))
        return Response(
            action=self.action, data=resp_json, raise_on_error=self._raise_on_error
        )

    # region #-- properties --#
    @property
    def action(self) -> str:
        """Return the action used in the request.

        :return: string containing the action
        """
        return self._action

    @property
    def payload(self) -> Optional[List[Dict] | Dict]:
        """Return the payload used for the request.

        :return: Optional[List[Dict] | Dict] containing the payload
        """
        return self._payload

    # endregion


class Response:
    """Represents a response from the API."""

    DATA_KEY_SINGLE: str = "output"
    DATA_KEY_TRANSACTION: str = "responses"
    RESULT_KEY: str = "result"

    def __init__(
        self, action: str, data: Dict[str, Any], raise_on_error: bool = True
    ) -> None:
        """Initialise the response.

        :param action: The action that was issued in the request to cause the response
        :param data: The JSON response received in response to the API call
        """
        self._action: str = action
        self._data: Dict[str, Any] = data
        self._log_formatter = Logger(prefix=f"{self.__class__.__name__}.")
        self._raise_on_error: bool = raise_on_error

        self._process_data()

    def _process_data(self) -> None:
        """Process the given data to check for errors."""
        if self._data.get(self.RESULT_KEY) != "OK" and self._raise_on_error:
            responses = self.data if self.action == Actions.TRANSACTION else [self.data]

            err = None
            for resp in responses:
                err = None
                if resp.get(self.RESULT_KEY) == "_ErrorInvalidInput":
                    err = MeshInvalidInput(resp.get("error"))
                elif resp.get(self.RESULT_KEY) == "_ErrorInvalidOutput":
                    err = MeshInvalidOutput(resp.get("error"))
                elif resp.get(self.RESULT_KEY) == "_ErrorUnauthorized":
                    err = MeshInvalidCredentials
                elif resp.get(self.RESULT_KEY) == "_ErrorUnknownAction":
                    action = (
                        resp.get("error")
                        if self.action == Actions.TRANSACTION
                        else f"Unknown action URI '{self.action}'"
                    )
                    err = MeshInvalidInput(action)
                elif resp.get(self.RESULT_KEY) == "ErrorDeviceNotInMasterMode":
                    err = MeshNodeNotPrimary
                elif resp.get(self.RESULT_KEY).startswith("_"):
                    err = MeshInvalidInput(
                        f"{resp.get(self.RESULT_KEY)}: '{self.action}'"
                    )

                if err:
                    break

            if err is None:
                _LOGGER.error(
                    self._log_formatter.format("unknown error received: %s"),
                    self._data,
                )
                err = MeshBadResponse

            raise err

    # region #-- properties --#
    @property
    def action(self) -> str:
        """Return the action that resulted in the response.

        :return: string containing the action
        """
        return self._action

    @property
    def data(self) -> Dict[str, Any]:
        """Return the response data."""
        ret = (
            self._data.get(self.DATA_KEY_TRANSACTION)
            if self.action == Actions.TRANSACTION
            else self._data.get(self.DATA_KEY_SINGLE, self._data)
        )

        return ret

    # endregion



"""pyvelop CLI."""

# region #-- imports --#
import logging
import uuid
from typing import Dict, List, Optional, Tuple

import aiohttp
import asyncclick as click

from pyvelop.const import _PACKAGE_NAME, _PACKAGE_VERSION
from pyvelop.device import Device
from pyvelop.exceptions import (
    MeshConnectionError,
    MeshDeviceNotFoundResponse,
    MeshInvalidCredentials,
    MeshInvalidInput,
    MeshNodeNotPrimary,
    MeshTimeoutError,
)
from pyvelop.mesh import Mesh
from pyvelop.node import Node

# endregion


class StandardCommand(click.Command):
    """Define standard options that should be used with all commands."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialise."""
        super().__init__(*args, **kwargs)

        def _create_session(ctx: click.Context, param: click.Option, value) -> None:
            """Create the session and store for late use."""
            if param.name == "create_session":
                if value:
                    _LOGGER.debug("Pre-creating a session")
                    ctx.obj = ctx.with_async_resource(
                        aiohttp.ClientSession(raise_for_status=True)
                    )

        def _setup_logging(_: click.Context, param: click.Option, value) -> None:
            """Handle logging."""
            if param.name == "verbose":
                if value:
                    logging.basicConfig()
                    _LOGGER.setLevel(logging.DEBUG)
                    _LOGGER.debug("Setting up logging")
                    if value > 1:
                        logging.getLogger(_PACKAGE_NAME).setLevel(logging.DEBUG)
                        logging.getLogger(f"{_PACKAGE_NAME}.jnap").setLevel(
                            logging.WARNING
                        )
                        logging.getLogger(f"{_PACKAGE_NAME}.jnap.verbose").setLevel(
                            logging.WARNING
                        )
                        logging.getLogger(f"{_PACKAGE_NAME}.mesh.verbose").setLevel(
                            logging.WARNING
                        )
                        if value > 2:
                            logging.getLogger(f"{_PACKAGE_NAME}.mesh.verbose").setLevel(
                                logging.DEBUG
                            )
                            if value > 3:
                                logging.getLogger(f"{_PACKAGE_NAME}.jnap").setLevel(
                                    logging.DEBUG
                                )
                                if value > 4:
                                    logging.getLogger(
                                        f"{_PACKAGE_NAME}.jnap.verbose"
                                    ).setLevel(logging.DEBUG)

        standard_options: List[click.Option] = [
            click.Option(
                ("-a", "--primary-node"),
                help="The primary node to direct all queries to.",
                required=True,
                type=str,
            ),
            click.Option(
                ("-c", "--create-session"),
                callback=_create_session,
                default=False,
                help="Supply this argument to create a session to pass into the library.",
                hidden=True,
                is_flag=True,
            ),
            click.Option(
                ("-p", "--password"),
                help="The local mesh password.",
                prompt=True,
                required=True,
            ),
            click.Option(
                ("-t", "--timeout"),
                default=30,
                help="The timeout for a request.",
                type=int,
            ),
            click.Option(
                ("-u", "--username"),
                default="admin",
                help="The username for communications.",
                type=str,
            ),
            click.Option(
                ("-v", "--verbose"),
                callback=_setup_logging,
                count=True,
                help="Set the verbosity of logging.",
            ),
        ]

        standard_options.reverse()
        for opt in standard_options:
            self.params.insert(0, opt)


DEF_INDENT: int = 2

click.anyio_backend = "aysncio"
_LOGGER = logging.getLogger(f"{_PACKAGE_NAME}.cli")


@click.group()
@click.version_option(version=_PACKAGE_VERSION, message="%(version)s")
def cli() -> None:
    """CLI for interacting with the pyvelop module."""


@cli.command(cls=StandardCommand)
@click.pass_context
@click.argument("device_name")
async def device(
    ctx: click.Context,
    device_name: str,
    **_,
) -> None:
    """Get details about a device on the Mesh."""
    if mesh_details := await mesh_connect(ctx):
        async with mesh_details:
            try:
                _ = uuid.UUID(device_name)
                device_details: List[Device] = [
                    await mesh_details.async_get_device_from_id(
                        device_id=device_name, force_refresh=True
                    )
                ]
            except ValueError:  # not a GUID
                device_details: List[Device] = await mesh_details.async_get_devices()

            for found_device in device_details:
                if (
                    found_device.name == device_name
                    or found_device.unique_id == device_name
                ):
                    _display_data(
                        _build_display_data(
                            mappings=[
                                ("results_time", "Queried at"),
                                ("unique_id", "Device ID"),
                                ("ui_type", "Icon Type"),
                                ("manufacturer", "Manufacturer"),
                                ("model", "Model"),
                                ("description", "Description"),
                                ("operating_system", "Operating System"),
                                ("serial", "Serial #"),
                                ("status", "Online"),
                                ("parent_name", "Parent"),
                                (
                                    "connected_adapters",
                                    "Connections",
                                    _connected_details(
                                        adapters=found_device.connected_adapters
                                    ),
                                ),
                                (
                                    "parental_control_schedule",
                                    "Parental Control",
                                    _parental_control_schedule_details(
                                        schedule=found_device.parental_control_schedule
                                    ),
                                ),
                            ],
                            obj=found_device,
                            title=found_device.name,
                        )
                    )
                    break


@cli.command(cls=StandardCommand)
@click.pass_context
async def mesh(
    ctx: click.Context,
    **_,
) -> None:
    """Get details about the Mesh."""
    indent: int = DEF_INDENT
    prefix: str = f"\n{indent * ' '}"
    if mesh_details := await mesh_connect(ctx):
        async with mesh_details:
            await mesh_details.async_gather_details()
            _display_data(
                _build_display_data(
                    mappings=[
                        ("wan_status", "Internet Connected"),
                        ("wan_ip", "Public IP"),
                        ("wan_dns", "DNS Servers", ", ".join(mesh_details.wan_dns)),
                        ("wan_mac", "MAC"),
                        (
                            "nodes",
                            "Nodes",
                            prefix
                            + prefix.join([node.name for node in mesh_details.nodes]),
                        ),
                        ("latest_speedtest_result", "Latest Speedtest Result"),
                        ("parental_control_enabled", "Parental Control Enabled"),
                        (
                            "guest_wifi_details",
                            "Guest Wi-Fi Details",
                            _guest_wifi_details(
                                state=mesh_details.guest_wifi_enabled,
                                networks=mesh_details.guest_wifi_details,
                            ),
                        ),
                        (
                            "storage_details",
                            "Storage Details",
                            _storage_details(
                                available_shares=mesh_details.storage_available,
                                server_details=mesh_details.storage_settings,
                            ),
                        ),
                        (
                            "devices",
                            f"Online Devices ({len([device for device in mesh_details.devices if device.status])})",
                            prefix
                            + prefix.join(
                                [
                                    f"{device.name} ({device.connected_adapters[0].get('ip')})"
                                    for device in mesh_details.devices
                                    if device.status
                                ]
                            ),
                        ),
                        (
                            "devices",
                            "Offline Devices "
                            f"({len([device for device in mesh_details.devices if not device.status])})",
                            prefix
                            + prefix.join(
                                [
                                    device.name
                                    for device in mesh_details.devices
                                    if not device.status
                                ]
                            ),
                        ),
                    ],
                    obj=mesh_details,
                    title="Mesh Overview",
                )
            )


@cli.group()
@click.help_option()
async def node() -> None:
    """Work with nodes on the Mesh."""


@node.command(cls=StandardCommand)
@click.argument("node_name")
@click.pass_context
async def details(
    ctx: click.Context,
    node_name: str,
    **_,
) -> None:
    """Get details about a node on the Mesh."""
    indent: int = DEF_INDENT
    prefix: str = f"\n{indent * ' '}"
    if mesh_details := await mesh_connect(ctx):
        async with mesh_details:
            await mesh_details.async_gather_details()
            node_details: List[Node] = mesh_details.nodes
            if not node_details:
                print("No nodes found")
            else:
                found_node: Node
                for found_node in node_details:
                    if found_node.name == node_name:
                        _display_data(
                            _build_display_data(
                                mappings=[
                                    ("results_time", "Queried at"),
                                    ("unique_id", "Device ID"),
                                    ("type", "Node type", found_node.type.title()),
                                    ("manufacturer", "Manufacturer"),
                                    ("model", "Model"),
                                    ("hardware_version", "Hardware version"),
                                    ("serial", "Serial #"),
                                    ("ui_type", "Icon type"),
                                    (
                                        "firmware",
                                        "Firmware",
                                        found_node.firmware.get("version"),
                                    ),
                                    (
                                        "firmware",
                                        "Latest firmware",
                                        found_node.firmware.get("latest_version"),
                                    ),
                                    ("last_update_check", "Last update check"),
                                    ("status", "Online"),
                                    (
                                        "connected_adapters",
                                        "Connections",
                                        _connected_details(
                                            adapters=found_node.connected_adapters
                                        ),
                                    ),
                                    (
                                        "backhaul",
                                        "Backhaul",
                                        "\n"
                                        + _build_display_data(
                                            indent=indent,
                                            mappings=[
                                                ("parent_name", "Parent"),
                                                ("connection", "Connection type"),
                                                (
                                                    "speed_mbps",
                                                    "Speed",
                                                    f"{found_node.backhaul.get('speed_mbps')}mbps"
                                                    if found_node.backhaul.get(
                                                        "speed_mbps"
                                                    )
                                                    else None,
                                                ),
                                                ("signal_strength", "Signal strength"),
                                                (
                                                    "rssi_dbm",
                                                    "RSSI",
                                                    f"{found_node.backhaul.get('rssi_dbm')}dBm"
                                                    if found_node.backhaul.get(
                                                        "rssi_dbm"
                                                    )
                                                    else None,
                                                ),
                                                ("last_checked", "Last checked"),
                                            ],
                                            obj=dict(
                                                **found_node.backhaul,
                                                parent_name=found_node.parent_name,
                                            ),
                                        ),
                                    ),
                                    (
                                        "connected_devices",
                                        f"Connected devices ({len(found_node.connected_devices)})",
                                        prefix
                                        + prefix.join(
                                            [
                                                device.get("name")
                                                for device in found_node.connected_devices
                                            ]
                                        ),
                                    ),
                                ],
                                obj=found_node,
                                title=node_name,
                            )
                        )
                        break


@node.command(cls=StandardCommand)
@click.argument("node_name")
@click.pass_context
async def restart(
    ctx: click.Context,
    node_name: str,
    **_,
) -> None:
    """Restart a node on the Mesh."""
    if mesh_details := await mesh_connect(ctx):
        async with mesh_details:
            try:
                await mesh_details.async_gather_details()
                _LOGGER.debug("Restarting %s", node_name)
                await mesh_details.async_reboot_node(node_name=node_name)
            except (MeshDeviceNotFoundResponse, MeshInvalidInput) as err:
                _LOGGER.error(err)


async def mesh_connect(ctx: click.Context = None) -> Optional[Mesh]:
    """Return the Mesh object."""
    if ctx is not None:
        mesh_object: Mesh = Mesh(
            node=ctx.params.get("primary_node"),
            password=ctx.params.get("password"),
            request_timeout=ctx.params.get("timeout"),
            session=await ctx.obj if ctx.obj else None,
            username=ctx.params.get("username"),
        )
        try:
            async with mesh_object:
                if not await mesh_object.async_test_credentials():
                    raise MeshInvalidCredentials
        except MeshConnectionError:
            _LOGGER.error("Unable to connect to %s", ctx.params.get("primary_node"))
        except MeshInvalidCredentials:
            _LOGGER.error(
                "Unable to authenticate with %s using provided credentials",
                ctx.params.get("primary_node"),
            )
        except MeshNodeNotPrimary:
            _LOGGER.error("%s is not the primary node", ctx.params.get("primary_node"))
        except MeshTimeoutError:
            _LOGGER.error("Timed out connecting to %s", ctx.params.get("primary_node"))
        else:
            return mesh_object

    return None


def _build_display_data(
    mappings: List[Tuple],
    obj: Device | Dict | Mesh | Node,
    indent: int = 0,
    title: str = "",
):
    """Build the string to display the given data."""
    ret: str = ""
    if title:
        ret = f"{title}\n"
        ret += f"{len(title) * '-'}\n"

    for properties in mappings:
        try:
            property_name, display_name, display_value = properties
        except ValueError:
            display_value = None
            property_name, display_name = properties

        if display_value is None:
            if isinstance(obj, Dict):
                display_value = obj.get(property_name)
            else:
                display_value = getattr(obj, property_name, None)

        ret += f"{indent * ' '}{display_name}: {display_value}\n"

    return ret.rstrip()


def _connected_details(adapters: List[Dict]) -> str:
    """Format the connected adapter details for display."""
    ret: str = ""
    indent: int = DEF_INDENT
    if not adapters:
        return "N/A"

    adapter = adapters[0]

    ret = _build_display_data(
        indent=indent,
        mappings=[
            ("mac", "MAC"),
            ("ip", "IPv4"),
            ("ipv6", "IPv6"),
            ("guest_network", "Guest"),
            (
                "signal_strength",
                "Signal Strength",
                f"{adapter.get('signal_strength', None)} ({adapter.get('rssi', None)}dBm)"
                if adapter.get("signal_strength", None)
                else "N/A",
            ),
        ],
        obj=adapter,
    )

    return "\n" + ret.rstrip()


def _display_data(message: str = "") -> None:
    """Display the given data on screen."""
    print(message)


def _guest_wifi_details(state: bool, networks: List[Dict]) -> str:
    """Format the Guest Wi-Fi details for display."""
    ret: str = ""
    indent: int = DEF_INDENT
    ret = _build_display_data(
        indent=indent,
        mappings=[
            ("state", "Enabled"),
            (
                "networks",
                "Networks",
                "\n"
                + "\n".join(
                    [
                        f"{indent * 2 * ' '}{idx}: {network.get('ssid')} ({network.get('band')})"
                        for idx, network in enumerate(networks)
                    ]
                ),
            ),
        ],
        obj={
            "state": state,
            "networks": networks,
        },
    )

    return "\n" + ret.rstrip()


def _parental_control_schedule_details(schedule: Dict) -> str:
    """Format the parental control schedule for display."""
    indent: int = DEF_INDENT
    prefix: str = f"\n{indent * 2 * ' '}"
    ret: str = _build_display_data(
        indent=indent,
        mappings=[
            (
                "blocked_internet_access",
                "Blocked Access",
                prefix
                + prefix.join(
                    [
                        f"{day.title()}: {', '.join(times) if times else 'N/A'}"
                        for day, times in schedule.get(
                            "blocked_internet_access", {}
                        ).items()
                    ]
                )
                if schedule.get("blocked_internet_access", {})
                else "N/A",
            ),
            (
                "blocked_sites",
                "Prohibited Sites",
                prefix + prefix.join(schedule.get("blocked_sites", []))
                if schedule.get("blocked_sites", [])
                else "N/A",
            ),
        ],
        obj=schedule,
    )

    return "\n" + ret.rstrip()


def _storage_details(available_shares: List[Dict], server_details: Dict) -> str:
    """Format the storage details for display."""
    ret: str = ""
    indent: int = DEF_INDENT

    def _build_share_data(share_details: Dict) -> str:
        """Format the share details for display."""
        return _build_display_data(
            indent=(indent * 3),
            mappings=[
                ("ip", "IP"),
                ("used_percent", "Used", f"{share_details.get('used_percent')}%"),
            ],
            obj=share_details,
        )

    ret = _build_display_data(
        indent=indent,
        mappings=[
            ("anonymous_access", "Anonymous Access"),
            (
                "available_shares",
                "Available Shares",
                f"\n{indent * 2 * ' '}"
                + f"\n{indent * 2 * ' '}".join(
                    [
                        f"{share.get('label')}:\n{_build_share_data(share_details=share)}"
                        for share in available_shares
                    ]
                ),
            ),
        ],
        obj=dict(server_details, available_shares=available_shares),
    )

    return "\n" + ret.rstrip()


if __name__ == "__main__":
    cli()



"""Exceptions for the pyvelop module."""


class MeshException(Exception):
    """Base Exsception for the Mesh."""


class MeshBadResponse(MeshException):
    """API returns a bad response."""

    def __init__(self) -> None:
        """Initialise and default message."""
        super().__init__("Bad Response")


class MeshConnectionError(MeshException):
    """Connection error for the API."""

    def __init__(self) -> None:
        """Initialise and default message."""
        super().__init__("Connection Error")


class MeshDeviceNotFoundResponse(MeshException):
    """Device is not found in the mesh."""

    def __init__(self) -> None:
        """Initialise and default message."""
        super().__init__("Device not found")


class MeshInvalidArguments(MeshException):
    """Invalid arguments have been passed to a function."""

    def __init__(self) -> None:
        """Initialise and default message."""
        super().__init__("Invalid Arguments")


class MeshInvalidCredentials(MeshException):
    """Credentials are invalid."""

    def __init__(self) -> None:
        """Initialise and default message."""
        super().__init__("Invalid Credentials")


class MeshInvalidInput(MeshException):
    """Parameters passed to the API are in valid."""


class MeshInvalidOutput(MeshException):
    """Invalid information would be returned from the API."""


class MeshNeedsGatherDetails(MeshException):
    """Must run the async_gather_details method first."""


class MeshNodeNotPrimary(MeshException):
    """API call being used on a node that isn't the primary."""

    def __init__(self) -> None:
        """Initialise and default message."""
        super().__init__("Node not Primary")


class MeshTimeoutError(MeshException):
    """Timeout error for the API."""

    def __init__(self) -> None:
        """Initialise and default message."""
        super().__init__("Timeout Error")


class MeshTooManyMatches(MeshException):
    """Too many matching devices when only one should be found."""

    def __init__(self) -> None:
        """Initialise and default message."""
        super().__init__("Too Many Matches")



"""Logging."""

# region #-- imports --#
import inspect

# endregion


class Logger:
    """Provide functions for managing log messages."""

    def __init__(self, unique_id: str = "", prefix: str = ""):
        """Initialise the formatter."""
        self._unique_id: str = unique_id
        self._prefix: str = prefix

    def format(self, msg: str, include_lineno: bool = False) -> str:
        """Format a log message in the correct format."""
        caller: inspect.FrameInfo = inspect.stack()[1]
        line_no = f" --> line: {caller.lineno}" if include_lineno else ""
        unique_id = f" ({self._unique_id})" if self._unique_id else ""
        return f"{self._prefix}{caller.function}{unique_id}{line_no} --> {msg}"



"""Representation of a mesh device."""

import datetime
from typing import List, Optional

from .base import MeshDevice


def _textualise_schedule(schedules: dict) -> dict:
    """Establish a textual version of the schedule.

    The schedule is stored in a string of 48 characters ('0's and '1's) representing 30-minute periods.
    '0' is blocked, '1' is allowed.

    :param schedules: dictionary representing the schedules as per the API
    :return: a dictionary showing the times that internet access is blocked and the sites that are also blocked
    """
    _pc_schedule_blocked: str = "0"
    _pc_schedule_unblocked: str = "1"

    ret: dict = {}

    for day, schedule in schedules.items():
        schedule_text: List = []
        schedule_list = list(schedule)
        start: datetime = None
        end: datetime = None
        for pos, minute in enumerate(schedule_list):
            if not start and minute == _pc_schedule_blocked:
                start: datetime = (
                    datetime.datetime.combine(
                        datetime.datetime.today(), datetime.datetime.min.time()
                    )
                    + datetime.timedelta(minutes=pos * 30)
                ).time()
                if end:
                    end = None
            if start and (
                minute == _pc_schedule_unblocked or pos == len(schedule_list) - 1
            ):
                if pos == len(schedule_list) - 1:
                    pos += 1
                end: datetime = (
                    datetime.datetime.combine(
                        datetime.datetime.today(), datetime.datetime.min.time()
                    )
                    + datetime.timedelta(minutes=pos * 30)
                ).time()
                schedule_text.append(
                    f"{start.strftime('%H:%M')}-{end.strftime('%H:%M')}"
                )
                if start:
                    start = None

        ret[day] = schedule_text

    return ret


class Device(MeshDevice):
    """Represents a user device in the mesh, i.e. not a node."""

    def __init__(self, **kwargs):
        """Initialise the Device.

        :param kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.__parent_name: Optional[str] = None

    @property
    def description(self) -> Optional[str]:
        """Get the description.

        :return: Device description as per Velop
        """
        return self._attribs.get("model", {}).get("description", None)

    @property
    def manufacturer(self) -> Optional[str]:
        """Get the manufacturer.

        :return: Manufacturer as found by the Velop
        """
        return self._get_user_property(
            name="userDeviceManufacturer"
        ) or self._attribs.get("model", {}).get("manufacturer", None)

    @property
    def model(self) -> Optional[str]:
        """Get the model.

        :return: Model as found by the Velop
        """
        return self._get_user_property(
            name="userDeviceModelNumber"
        ) or self._attribs.get("model", {}).get("modelNumber", None)

    @property
    def operating_system(self) -> Optional[str]:
        """Get the OS.

        :return: The OS as identified by the Velop
        """
        return self._get_user_property(name="userDeviceOS") or self._attribs.get(
            "unit", {}
        ).get("operatingSystem", None)

    @property
    def parental_control_schedule(self) -> dict:
        """Return the schedule of the parental controls for the device.

        An empty dictionary means that there are no parental controls in place
        """
        ret: dict = {}
        if self._attribs.get("parental_controls"):
            for rule in self._attribs.get("parental_controls"):
                ret = {
                    "blocked_internet_access": _textualise_schedule(
                        rule.get("wanSchedule", {})
                    ),
                    "blocked_sites": rule.get("blockedURLs", []),
                }

        return ret

    @property
    def parent_name(self) -> Optional[str]:
        """Name of the node the device is connected to."""
        return self.__parent_name

    @property
    def serial(self) -> Optional[str]:
        """Get the serial number."""
        return self._attribs.get("unit", {}).get("serialNumber", None)



"""Constants for the pyvelop module."""

# region #-- generic details --#
_PACKAGE_NAME: str = "pyvelop"
_PACKAGE_AUTHOR: str = "uvjim"
_PACKAGE_VERSION: str = "2022.9.4"
# endregion

DEF_REDACT: str = "**REDACTED**"



"""Setup."""

# region #-- imports --#
from setuptools import find_packages, setup

from pyvelop.const import _PACKAGE_AUTHOR, _PACKAGE_NAME, _PACKAGE_VERSION

# endregion

with open("README.rst", "r", encoding="utf-8") as f:
    long_description: str = f.read()

setup(
    name=_PACKAGE_NAME,
    version=_PACKAGE_VERSION,
    author=_PACKAGE_AUTHOR,
    description="A Python library for the Linksys Velop Mesh system",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/uvjim/pyvelop",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    install_requires=[
        "aiohttp",
    ],
)
